@@ -17,10 +17,6 @@
 
    Written by Torbjorn Granlund, David MacKenzie, and Jim Meyering. */
 
-#ifdef _AIX
- #pragma alloca
-#endif
-
 #include <config.h>
 #include <stdio.h>
 #include <sys/types.h>
@@ -32,11 +28,11 @@
 #include "backupfile.h"
 #include "copy.h"
 #include "cp-hash.h"
-#include "dirname.h"
 #include "error.h"
-#include "mmap-stack.h"
+#include "dirname.h"
 #include "path-concat.h"
 #include "quote.h"
+#include "utimens.h"
 
 #define ASSIGN_BASENAME_STRDUPA(Dest, File_name)	\
   do							\
@@ -51,7 +47,7 @@
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "cp"
 
-#define AUTHORS N_ ("Torbjorn Granlund, David MacKenzie, and Jim Meyering")
+#define AUTHORS "Torbjorn Granlund", "David MacKenzie", "Jim Meyering"
 
 #ifndef _POSIX_VERSION
 uid_t geteuid ();
@@ -120,6 +116,9 @@ static int const reply_vals[] =
   I_ALWAYS_YES, I_ALWAYS_NO, I_ASK_USER
 };
 
+/* The error code to return to the system. */
+static int exit_status = 0;
+
 static struct option const long_opts[] =
 {
   {"archive", no_argument, NULL, 'a'},
@@ -286,8 +285,7 @@ re_protect (const char *const_dst_path, int src_offset,
   char *src_path;		/* The source name in `dst_path'. */
   uid_t myeuid = geteuid ();
 
-  dst_path = (char *) alloca (strlen (const_dst_path) + 1);
-  strcpy (dst_path, const_dst_path);
+  ASSIGN_STRDUPA (dst_path, const_dst_path);
   src_path = dst_path + src_offset;
 
   for (p = attr_list; p; p = p->next)
@@ -309,16 +307,14 @@ re_protect (const char *const_dst_path, int src_offset,
 
       if (x->preserve_timestamps)
 	{
-	  struct utimbuf utb;
+	  struct timespec timespec[2];
 
-	  /* There's currently no interface to set file timestamps with
-	     better than 1-second resolution, so discard any fractional
-	     part of the source timestamp.  */
+	  timespec[0].tv_sec = src_sb.st_atime;
+	  timespec[0].tv_nsec = TIMESPEC_NS (src_sb.st_atim);
+	  timespec[1].tv_sec = src_sb.st_mtime;
+	  timespec[1].tv_nsec = TIMESPEC_NS (src_sb.st_mtim);
 
-	  utb.actime = src_sb.st_atime;
-	  utb.modtime = src_sb.st_mtime;
-
-	  if (utime (dst_path, &utb))
+	  if (utimens (dst_path, timespec))
 	    {
 	      error (0, errno, _("failed to preserve times for %s"),
 		     quote (dst_path));
@@ -386,13 +382,12 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
   char *dst_dirname;		/* Leading path of `dirpath'. */
   size_t dirlen;		/* Length of leading path of `dirpath'. */
 
-  dirpath = (char *) alloca (strlen (const_dirpath) + 1);
-  strcpy (dirpath, const_dirpath);
+  ASSIGN_STRDUPA (dirpath, const_dirpath);
 
   src = dirpath + src_offset;
 
   dirlen = dir_len (dirpath);
-  dst_dirname = (char *) alloca (dirlen + 1);
+  dst_dirname = alloca (dirlen + 1);
   memcpy (dst_dirname, dirpath, dirlen);
   dst_dirname[dirlen] = '\0';
 
@@ -411,8 +406,7 @@ make_path_private (const char *const_dirpath, int src_offset, int mode,
 	{
 	  /* Add this directory to the list of directories whose modes need
 	     fixing later. */
-	  struct dir_attr *new =
-	    (struct dir_attr *) xmalloc (sizeof (struct dir_attr));
+	  struct dir_attr *new = xmalloc (sizeof *new);
 	  new->slash_offset = slash - dirpath;
 	  new->next = *attr_list;
 	  *attr_list = new;
@@ -539,18 +533,18 @@ do_copy (int n_files, char **file, const char *target_directory,
 
   if (!dest_is_dir)
     {
-      if (target_directory)
+      if (target_directory || 1 < n_files)
 	{
-	  error (0, 0, _("%s: specified target is not a directory"),
-		 quote (dest));
-	  usage (EXIT_FAILURE);
-	}
+	  char const *msg;
+	  if (new_dst)
+	    msg = N_("%s: specified destination directory does not exist");
+	  else if (target_directory)
+	    msg = N_("%s: specified target is not a directory");
+	  else /* n_files > 1 */
+	    msg =
+	  N_("copying multiple files, but last argument %s is not a directory");
 
-      if (n_files > 1)
-	{
-	  error (0, 0,
-	 _("copying multiple files, but last argument %s is not a directory"),
-	     quote (dest));
+	  error (0, 0, _(msg), quote (dest));
 	  usage (EXIT_FAILURE);
 	}
     }
@@ -604,8 +598,8 @@ do_copy (int n_files, char **file, const char *target_directory,
 						  &attr_list, &new_dst,
 						  x->xstat);
 	    }
-  	  else
-  	    {
+	  else
+	    {
 	      char *arg_base;
 	      /* Append the last component of `arg' to `dest'.  */
 
@@ -620,7 +614,7 @@ do_copy (int n_files, char **file, const char *target_directory,
 	    {
 	      /* make_path_private failed, so don't even attempt the copy. */
 	      ret = 1;
-  	    }
+	    }
 	  else
 	    {
 	      int copy_into_self;
@@ -694,8 +688,7 @@ do_copy (int n_files, char **file, const char *target_directory,
 	  char *source_base;
 
 	  ASSIGN_BASENAME_STRDUPA (source_base, source);
-	  new_dest = (char *) alloca (strlen (dest)
-				      + strlen (source_base) + 1);
+	  new_dest = alloca (strlen (dest) + strlen (source_base) + 1);
 	  stpcpy (stpcpy (new_dest, dest), source_base);
 	}
       else
@@ -768,7 +761,7 @@ decode_preserve_arg (char const *arg, struct cp_options *x, int on_off)
       PRESERVE_OWNERSHIP, PRESERVE_LINK, PRESERVE_ALL
     };
 
-  /* Valid arguments to the `--reply' option. */
+  /* Valid arguments to the `--preserve' option. */
   static char const* const preserve_args[] =
     {
       "mode", "timestamps",
@@ -824,25 +817,6 @@ decode_preserve_arg (char const *arg, struct cp_options *x, int on_off)
   free (arg_writable);
 }
 
-static void run (size_t, char **, char const *, struct cp_options *)
-     ATTRIBUTE_NORETURN;
-/* Encapsulate the `copy-and-exit' functionality.  */
-static void
-run (size_t n_files, char **files, char const *target_directory,
-     struct cp_options *x)
-{
-  int fail;
-
-  /* Allocate space for remembering copied and created files.  */
-  hash_init ();
-
-  fail = do_copy (n_files, files, target_directory, x);
-
-  forget_all ();
-
-  exit (fail ? EXIT_FAILURE : EXIT_SUCCESS);
-}
-
 int
 main (int argc, char **argv)
 {
@@ -854,6 +828,7 @@ main (int argc, char **argv)
   int copy_contents = 0;
   char *target_directory = NULL;
 
+  initialize_main (&argc, &argv);
   program_name = argv[0];
   setlocale (LC_ALL, "");
   bindtextdomain (PACKAGE, LOCALEDIR);
@@ -1067,6 +1042,13 @@ main (int argc, char **argv)
   if (x.unlink_dest_after_failed_open && (x.hard_link || x.symbolic_link))
     x.unlink_dest_before_opening = 1;
 
-  RUN_WITH_BIG_STACK_4 (run, argc - optind, argv + optind,
-			target_directory, &x);
+  /* Allocate space for remembering copied and created files.  */
+
+  hash_init ();
+
+  exit_status |= do_copy (argc - optind, argv + optind, target_directory, &x);
+
+  forget_all ();
+
+  exit (exit_status);
 }
