@@ -37,7 +37,7 @@
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "mv"
 
-#define AUTHORS N_ ("Mike Parker, David MacKenzie, and Jim Meyering")
+#define AUTHORS "Mike Parker", "David MacKenzie", "Jim Meyering"
 
 /* Initial number of entries in each hash table entry's table of inodes.  */
 #define INITIAL_HASH_MODULE 100
@@ -96,9 +96,8 @@ static void
 rm_option_init (struct rm_options *x)
 {
   x->unlink_dirs = 0;
-
   x->ignore_missing_files = 0;
-
+  x->root_dev_ino = NULL;
   x->recursive = 1;
 
   /* Should we prompt for removal, too?  No.  Prompting for the `move'
@@ -321,9 +320,9 @@ Mandatory arguments to long options are mandatory for short options too.\n\
       --backup[=CONTROL]       make a backup of each existing destination file\n\
   -b                           like --backup but does not accept an argument\n\
   -f, --force                  do not prompt before overwriting\n\
-                                 equivalent to --reply=yes\n\
+                                 (equivalent to --reply=yes)\n\
   -i, --interactive            prompt before overwrite\n\
-                                 equivalent to --reply=query\n\
+                                 (equivalent to --reply=query)\n\
 "), stdout);
       fputs (_("\
       --reply={yes,no,query}   specify how to handle the prompt about an\n\
