@@ -31,7 +31,7 @@
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "pinky"
 
-#define AUTHORS N_ ("Joseph Arceneaux, David MacKenzie, and Kaveh Ghazi")
+#define AUTHORS "Joseph Arceneaux", "David MacKenzie", "Kaveh Ghazi"
 
 #ifndef MAXHOSTNAMELEN
 # define MAXHOSTNAMELEN 64
