@@ -51,7 +51,7 @@ struct rlimit { size_t rlim_cur; };
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "sort"
 
-#define AUTHORS N_ ("Mike Haertel and Paul Eggert")
+#define AUTHORS "Mike Haertel", "Paul Eggert"
 
 #if HAVE_LANGINFO_CODESET
 # include <langinfo.h>
@@ -498,10 +498,7 @@ static void
 add_temp_dir (char const *dir)
 {
   if (temp_dir_count == temp_dir_alloc)
-    {
-      temp_dir_alloc = temp_dir_alloc ? temp_dir_alloc * 2 : 16;
-      temp_dirs = xrealloc (temp_dirs, sizeof (temp_dirs) * temp_dir_alloc);
-    }
+    temp_dirs = x2nrealloc (temp_dirs, &temp_dir_alloc, sizeof *temp_dirs);
 
   temp_dirs[temp_dir_count++] = dir;
 }
@@ -573,7 +570,7 @@ inittables (void)
 	  name[j] = '\0';
 	}
       qsort ((void *) monthtab, MONTHS_PER_YEAR,
-	     sizeof (struct month), struct_month_cmp);
+	     sizeof *monthtab, struct_month_cmp);
     }
 #endif
 }
@@ -1053,10 +1050,7 @@ fillbuf (struct buffer *buf, register FILE *fp, char const *file)
 
       /* The current input line is too long to fit in the buffer.
 	 Double the buffer size and try again.  */
-      if (2 * buf->alloc < buf->alloc)
-	xalloc_die ();
-      buf->alloc *= 2;
-      buf->buf = xrealloc (buf->buf, buf->alloc);
+      buf->buf = x2nrealloc (buf->buf, &buf->alloc, sizeof *(buf->buf));
     }
 }
 
@@ -1133,7 +1127,7 @@ static int
 numcompare (register const char *a, register const char *b)
 {
   register int tmpa, tmpb, tmp;
-  register size_t loga, logb;
+  register size_t log_a, log_b;
 
   tmpa = *a;
   tmpb = *b;
@@ -1186,20 +1180,20 @@ numcompare (register const char *a, register const char *b)
 
       tmp = tmpb - tmpa;
 
-      for (loga = 0; ISDIGIT (tmpa); ++loga)
+      for (log_a = 0; ISDIGIT (tmpa); ++log_a)
 	do
 	  tmpa = *++a;
 	while (IS_THOUSANDS_SEP (tmpa));
 
-      for (logb = 0; ISDIGIT (tmpb); ++logb)
+      for (log_b = 0; ISDIGIT (tmpb); ++log_b)
 	do
 	  tmpb = *++b;
 	while (IS_THOUSANDS_SEP (tmpb));
 
-      if (loga != logb)
-	return loga < logb ? 1 : -1;
+      if (log_a != log_b)
+	return log_a < log_b ? 1 : -1;
 
-      if (!loga)
+      if (!log_a)
 	return 0;
 
       return tmp;
@@ -1248,20 +1242,20 @@ numcompare (register const char *a, register const char *b)
 
       tmp = tmpa - tmpb;
 
-      for (loga = 0; ISDIGIT (tmpa); ++loga)
+      for (log_a = 0; ISDIGIT (tmpa); ++log_a)
 	do
 	  tmpa = *++a;
 	while (IS_THOUSANDS_SEP (tmpa));
 
-      for (logb = 0; ISDIGIT (tmpb); ++logb)
+      for (log_b = 0; ISDIGIT (tmpb); ++log_b)
 	do
 	  tmpb = *++b;
 	while (IS_THOUSANDS_SEP (tmpb));
 
-      if (loga != logb)
-	return loga < logb ? -1 : 1;
+      if (log_a != log_b)
+	return log_a < log_b ? -1 : 1;
 
-      if (!loga)
+      if (!log_a)
 	return 0;
 
       return tmp;
@@ -1316,7 +1310,7 @@ getmonth (const char *s, size_t len)
   if (len == 0)
     return 0;
 
-  month = (char *) alloca (len + 1);
+  month = alloca (len + 1);
   for (i = 0; i < len; ++i)
     month[i] = fold_toupper[UCHAR (s[i])];
   len -= trailing_blanks (month, len);
@@ -1389,7 +1383,7 @@ keycompare (const struct line *a, const struct line *b)
 	{
 	  if (ignore || translate)
 	    {
-	      char *copy_a = (char *) alloca (lena + 1 + lenb + 1);
+	      char *copy_a = alloca (lena + 1 + lenb + 1);
 	      char *copy_b = copy_a + lena + 1;
 	      size_t new_len_a, new_len_b, i;
 
@@ -1992,7 +1986,8 @@ sort (char * const *files, int nfiles, char const *output_file)
       char const *file = *files;
       FILE *fp = xfopen (file, "r");
       FILE *tfp;
-      size_t bytes_per_line = 2 * sizeof (struct line) - sizeof (struct line) / 2;
+      size_t bytes_per_line = (2 * sizeof (struct line)
+			       - sizeof (struct line) / 2);
 
       if (! buf.alloc)
 	initbuf (&buf, bytes_per_line,
@@ -2059,7 +2054,7 @@ sort (char * const *files, int nfiles, char const *output_file)
     {
       int i = n_temp_files;
       struct tempnode *node;
-      char **tempfiles = xmalloc (n_temp_files * sizeof (char *));
+      char **tempfiles = xnmalloc (n_temp_files, sizeof *tempfiles);
       for (node = temphead; i > 0; node = node->next)
 	tempfiles[--i] = node->name;
       merge (tempfiles, n_temp_files, NMERGE, output_file);
@@ -2112,6 +2107,7 @@ parse_field_count (char const *string, size_t *val, char const *msgid)
 	break;
       /* Fall through.  */
     case LONGINT_OVERFLOW:
+    case LONGINT_OVERFLOW | LONGINT_INVALID_SUFFIX_CHAR:
       if (msgid)
 	error (SORT_FAILURE, 0, _("%s: count `%.*s' too large"),
 	       _(msgid), (int) (suffix - string), string);
@@ -2208,7 +2204,7 @@ set_ordering (register const char *s, struct keyfield *key,
 static struct keyfield *
 new_key (void)
 {
-  struct keyfield *key = xcalloc (1, sizeof *key);
+  struct keyfield *key = xzalloc (sizeof *key);
   key->eword = SIZE_MAX;
   return key;
 }
@@ -2311,7 +2307,7 @@ main (int argc, char **argv)
   gkey.numeric = gkey.general_numeric = gkey.month = gkey.reverse = false;
   gkey.skipsblanks = gkey.skipeblanks = false;
 
-  files = xmalloc (sizeof (char *) * argc);
+  files = xnmalloc (argc, sizeof *files);
 
   for (;;)
     {
