@@ -37,7 +37,7 @@
 /* The official name of this program (e.g., no `g' prefix).  */
 #define PROGRAM_NAME "who"
 
-#define AUTHORS N_ ("Joseph Arceneaux, David MacKenzie, and Michael Stone")
+#define AUTHORS "Joseph Arceneaux", "David MacKenzie", "Michael Stone"
 
 #ifndef MAXHOSTNAMELEN
 # define MAXHOSTNAMELEN 64
@@ -605,6 +605,8 @@ who (const char *filename)
     list_entries_who (n_users, utmp_buf);
   else
     scan_entries (n_users, utmp_buf);
+
+  free (utmp_buf);
 }
 
 void
